@@ -1,7 +1,7 @@
 /*
  *	The PCI Library -- ID to Name Translation
  *
- *	Copyright (c) 1997--2005 Martin Mares <mj@ucw.cz>
+ *	Copyright (c) 1997--2006 Martin Mares <mj@ucw.cz>
  *
  *	Can be freely distributed and used under the terms of the GNU GPL.
  */
@@ -14,11 +14,53 @@
 
 #include "internal.h"
 
+#ifdef PCI_COMPRESSED_IDS
+#include <zlib.h>
+typedef gzFile pci_file;
+#define pci_gets(f, l, s)	gzgets(f, l, s)
+#define pci_eof(f)		gzeof(f)
+
+static pci_file pci_open(struct pci_access *a)
+{
+  pci_file result;
+  size_t len;
+  char *new_name;
+
+  result = gzopen(a->id_file_name, "r");
+  if (result)
+    return result;
+  len = strlen(a->id_file_name);
+  if (len >= 3 && memcmp(a->id_file_name + len - 3, ".gz", 3) != 0)
+    return result;
+  new_name = malloc(len - 2);
+  memcpy(new_name, a->id_file_name, len - 3);
+  new_name[len - 3] = 0;
+  pci_set_name_list_path(a, new_name, 1);
+  return gzopen(a->id_file_name, "r");
+}
+
+#define pci_close(f)		gzclose(f)
+#define PCI_ERROR(f, err)						\
+	if (!err) {							\
+		int errnum;						\
+		err = gzerror(f, &errnum);				\
+		if (errnum == Z_ERRNO)	err = "I/O error";		\
+		else if (errnum >= 0)	err = NULL;			\
+	}
+#else
+typedef FILE * pci_file;
+#define pci_gets(f, l, s)	fgets(l, s, f)
+#define pci_eof(f)		feof(f)
+#define pci_open(a)		fopen(a->id_file_name, "r")
+#define pci_close(f)		fclose(f)
+#define PCI_ERROR(f, err)	if (!err && ferror(f))	err = "I/O error";
+#endif
+
 struct id_entry {
   struct id_entry *next;
   u32 id12, id34;
   byte cat;
-  byte name[1];
+  char name[1];
 };
 
 enum id_entry_type {
@@ -81,19 +123,21 @@ static inline unsigned int id_hash(int cat, u32 id12, u32 id34)
   return h % HASH_SIZE;
 }
 
-static struct id_entry *id_lookup(struct pci_access *a, int cat, int id1, int id2, int id3, int id4)
+static char *id_lookup(struct pci_access *a, int cat, int id1, int id2, int id3, int id4)
 {
   struct id_entry *n;
   u32 id12 = id_pair(id1, id2);
   u32 id34 = id_pair(id3, id4);
 
+  if (!a->id_hash)
+    return NULL;
   n = a->id_hash[id_hash(cat, id12, id34)];
   while (n && (n->id12 != id12 || n->id34 != id34 || n->cat != cat))
     n = n->next;
-  return n;
+  return n ? n->name : NULL;
 }
 
-static int id_insert(struct pci_access *a, int cat, int id1, int id2, int id3, int id4, byte *text)
+static int id_insert(struct pci_access *a, int cat, int id1, int id2, int id3, int id4, char *text)
 {
   u32 id12 = id_pair(id1, id2);
   u32 id34 = id_pair(id3, id4);
@@ -115,7 +159,7 @@ static int id_insert(struct pci_access *a, int cat, int id1, int id2, int id3, i
   return 0;
 }
 
-static int id_hex(byte *p, int cnt)
+static int id_hex(char *p, int cnt)
 {
   int x = 0;
   while (cnt--)
@@ -139,23 +183,23 @@ static inline int id_white_p(int c)
   return (c == ' ') || (c == '\t');
 }
 
-static const char *id_parse_list(struct pci_access *a, FILE *f, int *lino)
+static const char *id_parse_list(struct pci_access *a, pci_file f, int *lino)
 {
-  byte line[MAX_LINE];
-  byte *p;
+  char line[MAX_LINE];
+  char *p;
   int id1=0, id2=0, id3=0, id4=0;
   int cat = -1;
   int nest;
   static const char parse_error[] = "Parse error";
 
   *lino = 0;
-  while (fgets(line, sizeof(line), f))
+  while (pci_gets(f, line, sizeof(line)))
     {
       (*lino)++;
       p = line;
       while (*p && *p != '\n' && *p != '\r')
 	p++;
-      if (!*p && !feof(f))
+      if (!*p && !pci_eof(f))
 	return "Line too long";
       *p = 0;
       if (p > line && (p[-1] == ' ' || p[-1] == '\t'))
@@ -273,21 +317,22 @@ static const char *id_parse_list(struct pci_access *a, FILE *f, int *lino)
 int
 pci_load_name_list(struct pci_access *a)
 {
-  FILE *f;
+  pci_file f;
   int lino;
   const char *err;
 
   pci_free_name_list(a);
-  if (!(f = fopen(a->id_file_name, "r")))
+  a->hash_load_failed = 1;
+  if (!(f = pci_open(a)))
     return 0;
   a->id_hash = pci_malloc(a, sizeof(struct id_entry *) * HASH_SIZE);
-  bzero(a->id_hash, sizeof(struct id_entry *) * HASH_SIZE);
+  memset(a->id_hash, 0, sizeof(struct id_entry *) * HASH_SIZE);
   err = id_parse_list(a, f, &lino);
-  if (!err && ferror(f))
-    err = "I/O error";
-  fclose(f);
+  PCI_ERROR(f, err);
+  pci_close(f);
   if (err)
     a->error("%s at %s, line %d\n", err, a->id_file_name, lino);
+  a->hash_load_failed = 0;
   return 1;
 }
 
@@ -304,9 +349,10 @@ pci_free_name_list(struct pci_access *a)
     }
 }
 
-static struct id_entry *id_lookup_subsys(struct pci_access *a, int iv, int id, int isv, int isd)
+static char *
+id_lookup_subsys(struct pci_access *a, int iv, int id, int isv, int isd)
 {
-  struct id_entry *d = NULL;
+  char *d = NULL;
   if (iv > 0 && id > 0)						/* Per-device lookup */
     d = id_lookup(a, ID_SUBSYSTEM, iv, id, isv, isd);
   if (!d)							/* Generic lookup */
@@ -316,168 +362,158 @@ static struct id_entry *id_lookup_subsys(struct pci_access *a, int iv, int id, i
   return d;
 }
 
-char *
-pci_lookup_name(struct pci_access *a, char *buf, int size, int flags, ...)
+static char *
+format_name(char *buf, int size, int flags, char *name, char *num, char *unknown)
 {
-  va_list args;
-  int num, res, synth;
-  struct id_entry *v, *d, *cls, *pif;
-  int iv, id, isv, isd, icls, ipif;
+  int res;
+  if ((flags & PCI_LOOKUP_NO_NUMBERS) && !name)
+    return NULL;
+  else if (flags & PCI_LOOKUP_NUMERIC)
+    res = snprintf(buf, size, "%s", num);
+  else if (!name)
+    res = snprintf(buf, size, ((flags & PCI_LOOKUP_MIXED) ? "%s [%s]" : "%s %s"), unknown, num);
+  else if (!(flags & PCI_LOOKUP_MIXED))
+    res = snprintf(buf, size, "%s", name);
+  else
+    res = snprintf(buf, size, "%s [%s]", name, num);
+  if (res < 0 || res >= size)
+    return "<pci_lookup_name: buffer too small>";
+  else
+    return buf;
+}
 
-  va_start(args, flags);
-
-  num = 0;
-  if ((flags & PCI_LOOKUP_NUMERIC) || a->numeric_ids)
+static char *
+format_name_pair(char *buf, int size, int flags, char *v, char *d, char *num)
+{
+  int res;
+  if ((flags & PCI_LOOKUP_NO_NUMBERS) && (!v || !d))
+    return NULL;
+  if (flags & PCI_LOOKUP_NUMERIC)
+    res = snprintf(buf, size, "%s", num);
+  else if (flags & PCI_LOOKUP_MIXED)
     {
-      flags &= ~PCI_LOOKUP_NUMERIC;
-      num = 1;
-    }
-  else if (!a->id_hash)
-    {
-      if (!pci_load_name_list(a))
-	num = a->numeric_ids = 1;
-    }
-
-  if (flags & PCI_LOOKUP_NO_NUMBERS)
-    {
-      flags &= ~PCI_LOOKUP_NO_NUMBERS;
-      synth = 0;
-      if (num)
-	return NULL;
+      if (v && d)
+	res = snprintf(buf, size, "%s %s [%s]", v, d, num);
+      else if (!v)
+	res = snprintf(buf, size, "Unknown device [%s]", num);
+      else /* v && !d */
+	res = snprintf(buf, size, "%s Unknown device [%s]", v, num);
     }
   else
-    synth = 1;
-
-  switch (flags)
     {
-    case PCI_LOOKUP_VENDOR:
-      iv = va_arg(args, int);
-      if (num)
-	res = snprintf(buf, size, "%04x", iv);
-      else if (v = id_lookup(a, ID_VENDOR, iv, 0, 0, 0))
-	return v->name;
-      else
-	res = snprintf(buf, size, "Unknown vendor %04x", iv);
-      break;
-    case PCI_LOOKUP_DEVICE:
-      iv = va_arg(args, int);
-      id = va_arg(args, int);
-      if (num)
-	res = snprintf(buf, size, "%04x", id);
-      else if (d = id_lookup(a, ID_DEVICE, iv, id, 0, 0))
-	return d->name;
-      else if (synth)
-	res = snprintf(buf, size, "Unknown device %04x", id);
-      else
-	return NULL;
-      break;
-    case PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE:
-      iv = va_arg(args, int);
-      id = va_arg(args, int);
-      if (num)
-	res = snprintf(buf, size, "%04x:%04x", iv, id);
-      else
-	{
-	  v = id_lookup(a, ID_VENDOR, iv, 0, 0, 0);
-	  d = id_lookup(a, ID_DEVICE, iv, id, 0, 0);
-	  if (v && d)
-	    res = snprintf(buf, size, "%s %s", v->name, d->name);
-	  else if (!synth)
-	    return NULL;
-	  else if (!v)
-	    res = snprintf(buf, size, "Unknown device %04x:%04x", iv, id);
-	  else	/* !d */
-	    res = snprintf(buf, size, "%s Unknown device %04x", v->name, id);
-	}
-      break;
-    case PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_VENDOR:
-      isv = va_arg(args, int);
-      if (num)
-	res = snprintf(buf, size, "%04x", isv);
-      else if (v = id_lookup(a, ID_VENDOR, isv, 0, 0, 0))
-	return v->name;
-      else if (synth)
-	res = snprintf(buf, size, "Unknown vendor %04x", isv);
-      else
-	return NULL;
-      break;
-    case PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_DEVICE:
-      iv = va_arg(args, int);
-      id = va_arg(args, int);
-      isv = va_arg(args, int);
-      isd = va_arg(args, int);
-      if (num)
-	res = snprintf(buf, size, "%04x", isd);
-      else if (d = id_lookup_subsys(a, iv, id, isv, isd))
-	return d->name;
-      else if (synth)
-	res = snprintf(buf, size, "Unknown device %04x", isd);
-      else
-	return NULL;
-      break;
-    case PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE | PCI_LOOKUP_SUBSYSTEM:
-      iv = va_arg(args, int);
-      id = va_arg(args, int);
-      isv = va_arg(args, int);
-      isd = va_arg(args, int);
-      if (num)
-	res = snprintf(buf, size, "%04x:%04x", isv, isd);
-      else
-	{
-	  v = id_lookup(a, ID_VENDOR, isv, 0, 0, 0);
-	  d = id_lookup_subsys(a, iv, id, isv, isd);
-	  if (v && d)
-	    res = snprintf(buf, size, "%s %s", v->name, d->name);
-	  else if (!synth)
-	    return NULL;
-	  else if (!v)
-	    res = snprintf(buf, size, "Unknown device %04x:%04x", isv, isd);
-	  else /* !d */
-	    res = snprintf(buf, size, "%s Unknown device %04x", v->name, isd);
-	}
-      break;
-    case PCI_LOOKUP_CLASS:
-      icls = va_arg(args, int);
-      if (num)
-	res = snprintf(buf, size, "%04x", icls);
-      else if (cls = id_lookup(a, ID_SUBCLASS, icls >> 8, icls & 0xff, 0, 0))
-	return cls->name;
-      else if (cls = id_lookup(a, ID_CLASS, icls, 0, 0, 0))
-	res = snprintf(buf, size, "%s [%04x]", cls->name, icls);
-      else if (synth)
-	res = snprintf(buf, size, "Class %04x", icls);
-      else
-	return NULL;
-      break;
-    case PCI_LOOKUP_PROGIF:
-      icls = va_arg(args, int);
-      ipif = va_arg(args, int);
-      if (num)
-	res = snprintf(buf, size, "%02x", ipif);
-      else if (pif = id_lookup(a, ID_PROGIF, icls >> 8, icls & 0xff, ipif, 0))
-	return pif->name;
-      else if (icls == 0x0101 && !(ipif & 0x70))
-	{
-	  /* IDE controllers have complex prog-if semantics */
-	  res = snprintf(buf, size, "%s%s%s%s%s",
-			 (ipif & 0x80) ? "Master " : "",
-			 (ipif & 0x08) ? "SecP " : "",
-			 (ipif & 0x04) ? "SecO " : "",
-			 (ipif & 0x02) ? "PriP " : "",
-			 (ipif & 0x01) ? "PriO " : "");
-	  if (res > 0 && res < size)
-	    buf[--res] = 0;
-	}
-      else if (synth)
-	res = snprintf(buf, size, "ProgIf %02x", ipif);
-      else
-	return NULL;
-      break;
-    default:
-      return "<pci_lookup_name: invalid request>";
+      if (v && d)
+	res = snprintf(buf, size, "%s %s", v, d);
+      else if (!v)
+	res = snprintf(buf, size, "Unknown device %s", num);
+      else /* v && !d */
+	res = snprintf(buf, size, "%s Unknown device %s", v, num+5);
     }
   if (res < 0 || res >= size)
     return "<pci_lookup_name: buffer too small>";
   else
     return buf;
 }
+
+char *
+pci_lookup_name(struct pci_access *a, char *buf, int size, int flags, ...)
+{
+  va_list args;
+  char *v, *d, *cls, *pif;
+  int iv, id, isv, isd, icls, ipif;
+  char numbuf[16], pifbuf[32];
+
+  va_start(args, flags);
+
+  if (!(flags & PCI_LOOKUP_NO_NUMBERS))
+    {
+      if (a->numeric_ids > 1)
+	flags |= PCI_LOOKUP_MIXED;
+      else if (a->numeric_ids)
+	flags |= PCI_LOOKUP_NUMERIC;
+    }
+  if (flags & PCI_LOOKUP_MIXED)
+    flags &= ~PCI_LOOKUP_NUMERIC;
+
+  if (!a->id_hash && !(flags & PCI_LOOKUP_NUMERIC) && !a->hash_load_failed)
+    pci_load_name_list(a);
+
+  switch (flags & 0xffff)
+    {
+    case PCI_LOOKUP_VENDOR:
+      iv = va_arg(args, int);
+      sprintf(numbuf, "%04x", iv);
+      return format_name(buf, size, flags, id_lookup(a, ID_VENDOR, iv, 0, 0, 0), numbuf, "Unknown vendor");
+    case PCI_LOOKUP_DEVICE:
+      iv = va_arg(args, int);
+      id = va_arg(args, int);
+      sprintf(numbuf, "%04x", id);
+      return format_name(buf, size, flags, id_lookup(a, ID_DEVICE, iv, id, 0, 0), numbuf, "Unknown device");
+    case PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE:
+      iv = va_arg(args, int);
+      id = va_arg(args, int);
+      sprintf(numbuf, "%04x:%04x", iv, id);
+      v = id_lookup(a, ID_VENDOR, iv, 0, 0, 0);
+      d = id_lookup(a, ID_DEVICE, iv, id, 0, 0);
+      return format_name_pair(buf, size, flags, v, d, numbuf);
+    case PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_VENDOR:
+      isv = va_arg(args, int);
+      sprintf(numbuf, "%04x", isv);
+      v = id_lookup(a, ID_VENDOR, isv, 0, 0, 0);
+      return format_name(buf, size, flags, v, numbuf, "Unknown vendor");
+    case PCI_LOOKUP_SUBSYSTEM | PCI_LOOKUP_DEVICE:
+      iv = va_arg(args, int);
+      id = va_arg(args, int);
+      isv = va_arg(args, int);
+      isd = va_arg(args, int);
+      sprintf(numbuf, "%04x", isd);
+      return format_name(buf, size, flags, id_lookup_subsys(a, iv, id, isv, isd), numbuf, "Unknown device");
+    case PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE | PCI_LOOKUP_SUBSYSTEM:
+      iv = va_arg(args, int);
+      id = va_arg(args, int);
+      isv = va_arg(args, int);
+      isd = va_arg(args, int);
+      v = id_lookup(a, ID_VENDOR, isv, 0, 0, 0);
+      d = id_lookup_subsys(a, iv, id, isv, isd);
+      sprintf(numbuf, "%04x:%04x", isv, isd);
+      return format_name_pair(buf, size, flags, v, d, numbuf);
+    case PCI_LOOKUP_CLASS:
+      icls = va_arg(args, int);
+      sprintf(numbuf, "%04x", icls);
+      cls = id_lookup(a, ID_SUBCLASS, icls >> 8, icls & 0xff, 0, 0);
+      if (!cls && (cls = id_lookup(a, ID_CLASS, icls >> 8, 0, 0, 0)))
+	{
+	  if (!(flags & PCI_LOOKUP_NUMERIC)) /* Include full class number */
+	    flags |= PCI_LOOKUP_MIXED;
+	}
+      return format_name(buf, size, flags, cls, numbuf, ((flags & PCI_LOOKUP_MIXED) ? "Unknown class" : "Class"));
+    case PCI_LOOKUP_PROGIF:
+      icls = va_arg(args, int);
+      ipif = va_arg(args, int);
+      sprintf(numbuf, "%02x", ipif);
+      pif = id_lookup(a, ID_PROGIF, icls >> 8, icls & 0xff, ipif, 0);
+      if (!pif && icls == 0x0101 && !(ipif & 0x70))
+	{
+	  /* IDE controllers have complex prog-if semantics */
+	  sprintf(pifbuf, "%s%s%s%s%s",
+		  (ipif & 0x80) ? " Master" : "",
+		  (ipif & 0x08) ? " SecP" : "",
+		  (ipif & 0x04) ? " SecO" : "",
+		  (ipif & 0x02) ? " PriP" : "",
+		  (ipif & 0x01) ? " PriO" : "");
+	  pif = pifbuf;
+	  if (*pif)
+	    pif++;
+	}
+      return format_name(buf, size, flags, pif, numbuf, "ProgIf");
+    default:
+      return "<pci_lookup_name: invalid request>";
+    }
+}
+
+void pci_set_name_list_path(struct pci_access *a, char *name, int to_be_freed)
+{
+  if (a->free_id_name)
+    free(a->id_file_name);
+  a->id_file_name = name;
+  a->free_id_name = to_be_freed;
+}
